# -*- coding: utf-8 -*-
#
'''
I/O for VTK, VTU, Exodus etc.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
import numpy


def read(filetype, filename):
    import vtk
    from vtk.util import numpy_support

    def _read_data(data):
        '''Extract numpy arrays from a VTK data set.
        '''
        # Go through all arrays, fetch data.
        out = {}
        for k in range(data.GetNumberOfArrays()):
            array = data.GetArray(k)
            if array:
                array_name = array.GetName()
                out[array_name] = vtk.util.numpy_support.vtk_to_numpy(array)

        return out

    def _read_cells(vtk_mesh):
        data = vtk.util.numpy_support.vtk_to_numpy(
                vtk_mesh.GetCells().GetData()
                )
        offsets = vtk.util.numpy_support.vtk_to_numpy(
                vtk_mesh.GetCellLocationsArray()
                )
        types = vtk.util.numpy_support.vtk_to_numpy(
                vtk_mesh.GetCellTypesArray()
                )

        vtk_to_meshio_type = {
            vtk.VTK_VERTEX: 'vertex',
            vtk.VTK_LINE: 'line',
            vtk.VTK_TRIANGLE: 'triangle',
            vtk.VTK_QUAD: 'quad',
            vtk.VTK_TETRA: 'tetra',
            vtk.VTK_HEXAHEDRON: 'hexahedron',
            vtk.VTK_WEDGE: 'wedge',
            vtk.VTK_PYRAMID: 'pyramid'
            }

        # `data` is a one-dimensional vector with
        # (num_points0, p0, p1, ... ,pk, numpoints1, p10, p11, ..., p1k, ...
        # Translate it into the cells dictionary.
        cells = {}
        for vtk_type, meshio_type in vtk_to_meshio_type.items():
            # Get all offsets for vtk_type
            os = offsets[numpy.argwhere(types == vtk_type).transpose()[0]]
            num_cells = len(os)
            if num_cells > 0:
                num_pts = data[os[0]]
                # instantiate the array
                arr = numpy.empty((num_cells, num_pts), dtype=int)
                # sort the num_pts entries after the offsets into the columns
                # of arr
                for k in range(num_pts):
                    arr[:, k] = data[os+k+1]
                cells[meshio_type] = arr

        return cells

    if filetype == 'vtk':
        reader = vtk.vtkUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()
        vtk_mesh = reader.GetOutput()
    elif filetype == 'vtu':
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()
        vtk_mesh = reader.GetOutput()
    elif filetype == 'xdmf':
        reader = vtk.vtkXdmfReader()
        reader.SetFileName(filename)
        reader.Update()
        vtk_mesh = reader.GetOutputDataObject(0)
    elif filetype == 'exodus':
        reader = vtk.vtkExodusIIReader()
        reader.SetFileName(filename)
        vtk_mesh = _read_exodusii_mesh(reader)
    else:
        raise RuntimeError('Unknown file type \'%s\'.' % filename)

    # Explicitly extract points, cells, point data, field data
    points = vtk.util.numpy_support.vtk_to_numpy(
            vtk_mesh.GetPoints().GetData()
            )
    cells = _read_cells(vtk_mesh)
    point_data = _read_data(vtk_mesh.GetPointData())
    field_data = _read_data(vtk_mesh.GetFieldData())

    cell_data = _read_data(vtk_mesh.GetCellData())
    # split cell_data by the cell type
    cd = {}
    index = 0
    for cell_type in cells:
        num_cells = len(cells[cell_type])
        cd[cell_type] = {}
        for name, array in cell_data.iteritems():
            cd[cell_type][name] = array[index:index+num_cells]
        index += num_cells
    cell_data = cd

    return points, cells, point_data, cell_data, field_data


# def _read_exodus_mesh(reader, file_name):
#     '''Uses a vtkExodusIIReader to return a vtkUnstructuredGrid.
#     '''
#     reader.SetFileName(file_name)
#
#     # Create Exodus metadata that can be used later when writing the file.
#     reader.ExodusModelMetadataOn()
#
#     # Fetch metadata.
#     reader.UpdateInformation()
#
#     # Make sure the point fields are read during Update().
#     for k in range(reader.GetNumberOfPointArrays()):
#         arr_name = reader.GetPointArrayName(k)
#         reader.SetPointArrayStatus(arr_name, 1)
#
#     # Read the file.
#     reader.Update()
#
#     return reader.GetOutput()


def _read_exodusii_mesh(reader, timestep=None):
    '''Uses a vtkExodusIIReader to return a vtkUnstructuredGrid.
    '''
    # Fetch metadata.
    reader.UpdateInformation()

    # Set time step to read.
    if timestep:
        reader.SetTimeStep(timestep)

    # Make sure the point fields are read during Update().
    for k in range(reader.GetNumberOfPointResultArrays()):
        arr_name = reader.GetPointResultArrayName(k)
        reader.SetPointResultArrayStatus(arr_name, 1)

    # Make sure the point fields are read during Update().
    for k in range(reader.GetNumberOfElementResultArrays()):
        arr_name = reader.GetElementResultArrayName(k)
        reader.SetElementResultArrayStatus(arr_name, 1)

    # Make sure all field data is read.
    for k in range(reader.GetNumberOfGlobalResultArrays()):
        arr_name = reader.GetGlobalResultArrayName(k)
        reader.SetGlobalResultArrayStatus(arr_name, 1)

    # Read the file.
    reader.Update()
    out = reader.GetOutput()

    # Loop through the blocks and search for a vtkUnstructuredGrid.
    # In Exodus, different element types are stored different meshes, with
    # point information possibly duplicated.
    vtk_mesh = []
    for i in range(out.GetNumberOfBlocks()):
        blk = out.GetBlock(i)
        for j in range(blk.GetNumberOfBlocks()):
            sub_block = blk.GetBlock(j)
            if sub_block.IsA('vtkUnstructuredGrid'):
                vtk_mesh.append(sub_block)

    if len(vtk_mesh) == 0:
        raise IOError('No \'vtkUnstructuredGrid\' found!')
    elif len(vtk_mesh) > 1:
        raise IOError('More than one \'vtkUnstructuredGrid\' found!')

    # Cut off trailing '_' from array names.
    for k in range(vtk_mesh[0].GetPointData().GetNumberOfArrays()):
        array = vtk_mesh[0].GetPointData().GetArray(k)
        array_name = array.GetName()
        if array_name[-1] == '_':
            array.SetName(array_name[0:-1])

    # time_values = reader.GetOutputInformation(0).Get(
    #     vtkStreamingDemandDrivenPipeline.TIME_STEPS()
    #     )

    return vtk_mesh[0]  # , time_values


def write(filetype,
          filename,
          points,
          cells,
          point_data=None,
          cell_data=None,
          field_data=None
          ):
    import vtk
    from vtk.util import numpy_support

    def _create_vtkarray(X, name):
        array = vtk.util.numpy_support.numpy_to_vtk(X, deep=1)
        array.SetName(name)
        return array

    if point_data is None:
        point_data = {}
    if cell_data is None:
        cell_data = {}
    if field_data is None:
        field_data = {}

    vtk_mesh = _generate_vtk_mesh(points, cells)

    # add point data
    pd = vtk_mesh.GetPointData()
    for name, X in point_data.iteritems():
        # There is a naming inconsistency in VTK when it comes to multivectors
        # in Exodus files:
        # If a vector 'v' has two components, they are called 'v_x', 'v_y'
        # (note the underscore), if it has three, then they are called 'vx',
        # 'vy', 'vz'. See bug <http://www.vtk.org/Bug/view.php?id=15894>.
        # For VT{K,U} files, no underscore is ever added.
        pd.AddArray(_create_vtkarray(X, name))

    # Add cell data.
    # The cell_data is structured like
    #
    #  cell_type ->
    #      key -> array
    #      key -> array
    #      [...]
    #  cell_type ->
    #      key -> array
    #      key -> array
    #      [...]
    #  [...]
    #
    # VTK expects one array for each `key`, so assemble the keys across all
    # mesh_types. This requires each key to be present for each mesh_type, of
    # course.
    all_keys = []
    for cell_type in cell_data:
        all_keys += cell_data[cell_type].keys()
    # create unified cell data
    for key in all_keys:
        for cell_type in cell_data:
            assert key in cell_data[cell_type]
    unified_cell_data = {
        key: numpy.concatenate([
            cell_data[cell_type][key]
            for cell_type in cell_data
            ])
        for key in all_keys
        }
    # add the array data to the mesh
    cd = vtk_mesh.GetCellData()
    for name, array in unified_cell_data.iteritems():
        cd.AddArray(_create_vtkarray(array, name))

    # add field data
    fd = vtk_mesh.GetFieldData()
    for key, value in field_data.iteritems():
        fd.AddArray(_create_vtkarray(value, key))

    if filetype in 'vtk-ascii':
        writer = vtk.vtkUnstructuredGridWriter()
        writer.SetFileTypeToASCII()
    elif filetype == 'vtk-binary':
        writer = vtk.vtkUnstructuredGridWriter()
        writer.SetFileTypeToBinary()
    elif filetype == 'vtu':  # vtk xml format
        writer = vtk.vtkXMLUnstructuredGridWriter()
    elif filetype == 'xdmf':
        writer = vtk.vtkXdmfWriter()
    elif filetype == 'exodus':   # exodus ii format
        writer = vtk.vtkExodusIIWriter()
        # if the mesh contains vtkmodeldata information, make use of it
        # and write out all time steps.
        writer.WriteAllTimeStepsOn()
    else:
        raise RuntimeError('unknown file type \'%s\'.' % filetype)

    writer.SetFileName(filename)
    try:
        writer.SetInput(vtk_mesh)
    except AttributeError:
        writer.SetInputData(vtk_mesh)
    writer.Write()

    return


def _generate_vtk_mesh(points, cells):
    import vtk
    from vtk.util import numpy_support

    mesh = vtk.vtkUnstructuredGrid()

    # set points
    vtk_points = vtk.vtkPoints()
    # Not using a deep copy here results in a segfault.
    vtk_array = numpy_support.numpy_to_vtk(points, deep=True)
    vtk_points.SetData(vtk_array)
    mesh.SetPoints(vtk_points)

    # Set cells.
    meshio_to_vtk_type = {
        'vertex': vtk.VTK_VERTEX,
        'line': vtk.VTK_LINE,
        'triangle': vtk.VTK_TRIANGLE,
        'quad': vtk.VTK_QUAD,
        'tetra': vtk.VTK_TETRA,
        'hexahedron': vtk.VTK_HEXAHEDRON,
        'wedge': vtk.VTK_WEDGE,
        'pyramid': vtk.VTK_PYRAMID
        }

    # create cell_array. It's a one-dimensional vector with
    # (num_points2, p0, p1, ... ,pk, numpoints1, p10, p11, ..., p1k, ...
    cell_types = []
    cell_offsets = []
    cell_connectivity = []
    len_array = 0
    for meshio_type, data in cells.iteritems():
        numcells, num_local_nodes = data.shape
        vtk_type = meshio_to_vtk_type[meshio_type]
        # add cell types
        cell_types.append(numpy.empty(numcells, dtype=numpy.ubyte))
        cell_types[-1].fill(vtk_type)
        # add cell offsets
        cell_offsets.append(numpy.arange(
            len_array,
            len_array + numcells * (num_local_nodes + 1),
            num_local_nodes + 1,
            dtype=numpy.int64
            ))
        cell_connectivity.append(
            numpy.c_[
                num_local_nodes * numpy.ones(numcells, dtype=data.dtype),
                data
            ].flatten()
            )
        len_array += len(cell_connectivity[-1])

    cell_types = numpy.concatenate(cell_types)
    cell_offsets = numpy.concatenate(cell_offsets)
    cell_connectivity = numpy.concatenate(cell_connectivity)

    connectivity = vtk.util.numpy_support.numpy_to_vtkIdTypeArray(
        cell_connectivity.astype(numpy.int64),
        deep=1
        )

    # wrap the data into a vtkCellArray
    cell_array = vtk.vtkCellArray()
    cell_array.SetCells(len(cell_types), connectivity)

    # Add cell data to the mesh
    mesh.SetCells(
        numpy_support.numpy_to_vtk(
            cell_types,
            deep=1,
            array_type=vtk.vtkUnsignedCharArray().GetDataType()
            ),
        numpy_support.numpy_to_vtk(
            cell_offsets,
            deep=1,
            array_type=vtk.vtkIdTypeArray().GetDataType()
            ),
        cell_array
        )

    return mesh



# -*- coding: utf-8 -*-
#
'''
I/O for Medit's format, cf.
<https://people.sc.fsu.edu/~jburkardt/data/medit/medit.html>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
from itertools import islice
import numpy


def read(filename):
    with open(filename) as f:
        points, cells = read_buffer(f)

    return points, cells, {}, {}, {}


def read_buffer(f):
    dim = 0
    cells = {}

    while True:
        try:
            line = next(islice(f, 1))
        except StopIteration:
            break

        stripped = line.strip()

        # skip comments and empty lines
        if len(stripped) == 0 or stripped[0] == '#':
            continue

        assert stripped[0].isalpha()
        keyword = stripped.split(' ')[0]

        meshio_from_medit = {
            'Edges': ('line', 2),
            'Triangles': ('triangle', 3),
            'Quadrilaterals': ('quad', 4),
            'Tetrahedra': ('tetra', 4),
            'Hexahedra': ('hexahedra', 8)
            }

        if keyword == 'MeshVersionFormatted':
            assert stripped[-1] == '1'
        elif keyword == 'Dimension':
            dim = int(stripped[-1])
        elif keyword == 'Vertices':
            assert dim > 0
            # The first line is the number of nodes
            line = next(islice(f, 1))
            num_verts = int(line)
            points = numpy.empty((num_verts, dim), dtype=float)
            for k, line in enumerate(islice(f, num_verts)):
                # Throw away the label immediately
                points[k] = numpy.array(line.split(), dtype=float)[:-1]
        elif keyword in meshio_from_medit:
            meshio_name, num = meshio_from_medit[keyword]
            # The first line is the number of elements
            line = next(islice(f, 1))
            num_cells = int(line)
            cell_data = numpy.empty((num_cells, num), dtype=int)
            for k, line in enumerate(islice(f, num_cells)):
                data = numpy.array(line.split(), dtype=int)
                # Throw away the label
                cell_data[k] = data[:-1]

            # adapt 0-base
            cells[meshio_name] = cell_data - 1
        elif keyword == 'End':
            pass
        else:
            raise RuntimeError('Unknown keyword \'%s\'.' % keyword)

    return points, cells


def write(
        filename,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None
        ):
    with open(filename, 'w') as fh:
        fh.write('MeshVersionFormatted 1\n')
        fh.write('# Created by meshio\n')

        # Dimension info
        fh.write('\nDimension %d\n' % points.shape[1])

        # vertices
        fh.write('\nVertices\n')
        fh.write('%d\n' % len(points))
        labels = numpy.ones(len(points), dtype=int)
        data = numpy.c_[points, labels]
        fmt = ' '.join(['%e'] * points.shape[1]) + ' %d'
        numpy.savetxt(fh, data, fmt)

        medit_from_meshio = {
            'line': ('Edges', 2),
            'triangle': ('Triangles', 3),
            'quad': ('Quadrilaterals', 4),
            'tetra': ('Tetrahedra', 4),
            'hexahedra': ('Hexahedra', 8)
            }

        for key, data in cells.items():
            medit_name, num = medit_from_meshio[key]
            fh.write('\n')
            fh.write('%s\n' % medit_name)
            fh.write('%d\n' % len(data))
            labels = numpy.ones(len(data), dtype=int)
            # adapt 1-base
            data_with_label = numpy.c_[data + 1, labels]
            fmt = ' '.join(['%d'] * (num + 1))
            numpy.savetxt(fh, data_with_label, fmt)

        fh.write('\nEnd\n')

    return



# -*- coding: utf-8 -*-
#
'''
I/O for Gmsh's msh format, cf.
<http://geuz.org/gmsh/doc/texinfo/gmsh.html#File-formats>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
from itertools import islice
import numpy
import warnings


def read(filename):
    '''Reads a Gmsh msh file.
    '''
    with open(filename) as f:
        points, cells, point_data, cell_data, field_data = read_buffer(f)

    return points, cells, point_data, cell_data, field_data


def read_buffer(f):
    # The format is specified at
    # <http://geuz.org/gmsh/doc/texinfo/gmsh.html#MSH-ASCII-file-format>.

    # Initialize the data optional data fields
    field_data = {}
    cell_data = {}
    point_data = {}

    while True:
        try:
            line = next(islice(f, 1))
        except StopIteration:
            break
        assert line[0] == '$'
        environ = line[1:].strip()
        if environ == 'MeshFormat':
            line = next(islice(f, 1))
            # 2.2 0 8
            line = next(islice(f, 1))
            assert line.strip() == '$EndMeshFormat'
        elif environ == 'PhysicalNames':
            line = next(islice(f, 1))
            num_phys_names = int(line)
            for k, line in enumerate(islice(f, num_phys_names)):
                key = line.split(' ')[2].replace('"', '').replace('\n', '')
                phys_group = int(line.split(' ')[1])
                field_data[key] = phys_group
            line = next(islice(f, 1))
            assert line.strip() == '$EndPhysicalNames'
        elif environ == 'Nodes':
            # The first line is the number of nodes
            # TODO speed up with <http://stackoverflow.com/a/41642053/353337>
            line = next(islice(f, 1))
            num_nodes = int(line)
            points = numpy.empty((num_nodes, 3))
            for k, line in enumerate(islice(f, num_nodes)):
                # Throw away the index immediately
                points[k, :] = numpy.array(line.split(), dtype=float)[1:]
            line = next(islice(f, 1))
            assert line.strip() == '$EndNodes'
        elif environ == 'Elements':
            # The first line is the number of elements
            line = next(islice(f, 1))
            num_cells = int(line)
            cells = {}
            gmsh_to_meshio_type = {
                    15: ('vertex', 1),
                    1: ('line', 2),
                    2: ('triangle', 3),
                    3: ('quad', 4),
                    4: ('tetra', 4),
                    5: ('hexahedron', 8),
                    6: ('wedge', 6),
                    7: ('pyramid', 5),
                    8: ('line3', 3),
                    9: ('triangle6', 6),
                    10: ('quad9', 9),
                    11: ('tetra10', 10),
                    12: ('hexahedron27', 27),
                    13: ('prism18', 18),
                    14: ('pyramid14', 14),
                    26: ('line4', 4),
                    36: ('quad16', 16)
                    }
            # For each cell, there are at least two tags: the physical entity
            # and the elementary geometrical entity the cell belongs to (see
            # <http://gmsh.info/doc/texinfo/gmsh.html#MSH-ASCII-file-format>).
            has_additional_tag_data = False
            for k, line in enumerate(islice(f, num_cells)):
                # Throw away the index (data[0]) immediately;
                data = numpy.array(line.split(), dtype=int)
                t = gmsh_to_meshio_type[data[1]]

                if t[0] not in cells:
                    cells[t[0]] = []

                # Subtract one to account for the fact that python indices
                # are 0-based.
                cells[t[0]].append(data[-t[1]:] - 1)

                # data[2] gives the number of tags. The gmsh manual
                # <http://gmsh.info/doc/texinfo/gmsh.html#MSH-ASCII-file-format>
                # says:
                # >>>
                # By default, the first tag is the number of the physical
                # entity to which the element belongs; the second is the number
                # of the elementary geometrical entity to which the element
                # belongs; the third is the number of mesh partitions to which
                # the element belongs, followed by the partition ids (negative
                # partition ids indicate ghost cells). A zero tag is equivalent
                # to no tag. Gmsh and most codes using the MSH 2 format require
                # at least the first two tags (physical and elementary tags).
                # <<<

                # Just remember if there's any more tag data so we can warn. No
                # idea how to store the data in an array yet since it can be
                # "sparse", i.e., not all cells need to have the same number of
                # tags.
                if data[2] >= 2:
                    if data[2] > 2:
                        has_additional_tag_data = True

                    if t[0] not in cell_data:
                        cell_data[t[0]] = {
                            'physical': [],
                            'geometrical': [],
                            }
                    cell_data[t[0]]['physical'].append(data[3])
                    cell_data[t[0]]['geometrical'].append(data[4])
                else:
                    # TODO handle data[2] == 1
                    assert data[2] == 0

            line = next(islice(f, 1))
            assert line.strip() == '$EndElements'
        else:
            raise RuntimeError('Unknown environment \'%s\'.' % environ)

    for key in cells:
        cells[key] = numpy.vstack(cells[key])

    if has_additional_tag_data:
        warnings.warn(
            'The file contains tag data that couldn\'t be processed.'
            )

    return points, cells, point_data, cell_data, field_data


def write(
        filename,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None
        ):
    '''Writes msh files, cf.
    http://geuz.org/gmsh/doc/texinfo/gmsh.html#MSH-ASCII-file-format
    '''
    if point_data is None:
        point_data = {}
    if cell_data is None:
        cell_data = {}
    if field_data is None:
        field_data = {}

    with open(filename, 'w') as fh:
        fh.write('$MeshFormat\n2 0 8\n$EndMeshFormat\n')

        # Write nodes
        fh.write('$Nodes\n')
        fh.write('%d\n' % len(points))
        for k, x in enumerate(points):
            fh.write('%d %f %f %f\n' % (k+1, x[0], x[1], x[2]))
        fh.write('$EndNodes\n')

        # Translate meshio types to gmsh codes
        # http://geuz.org/gmsh/doc/texinfo/gmsh.html#MSH-ASCII-file-format
        meshio_to_gmsh_type = {
                'vertex': 15,
                'line': 1,
                'triangle': 2,
                'quad': 3,
                'tetra': 4,
                'hexahedron': 5,
                'wedge': 6,
                }
        fh.write('$Elements\n')
        # count all cells
        num_cells = 0
        for key, data in cells.iteritems():
            num_cells += data.shape[0]
        fh.write('%d\n' % num_cells)

        consecutive_index = 0
        for cell_type, node_idcs in cells.iteritems():
            if cell_type in cell_data and len(cell_data[cell_type]) > 0:
                for key in cell_data[cell_type]:
                    # assert data consistency
                    assert len(cell_data[cell_type][key]) == len(node_idcs)
                    # TODO assert that the data type is int

                # if a tag is present, make sure that there are 'physical' and
                # 'geometrical' as well.
                if 'physical' not in cell_data[cell_type]:
                    cell_data[cell_type]['physical'] = \
                        numpy.ones(len(node_idcs))
                if 'geometrical' not in cell_data[cell_type]:
                    cell_data[cell_type]['geometrical'] = \
                        numpy.ones(len(node_idcs))

                # 'physical' and 'geometrical' go first; this is what the gmsh
                # file format prescribes
                keywords = cell_data[cell_type].keys()
                keywords.remove('physical')
                keywords.remove('geometrical')
                sorted_keywords = ['physical', 'geometrical'] + keywords
                fcd = numpy.column_stack([
                        cell_data[cell_type][key] for key in sorted_keywords
                        ])

                num_nodes_per_cell = node_idcs.shape[1]
                form = '%d ' + '%d' % meshio_to_gmsh_type[cell_type] \
                    + ' %d' % len(cell_data) + ' %d' * len(cell_data) \
                    + ' ' + ' '.join(num_nodes_per_cell * ['%d']) \
                    + '\n'
                for k, c in enumerate(node_idcs):
                    fh.write(
                        form % (
                            (consecutive_index + k + 1,) +
                            tuple(fcd[k]) +
                            tuple(c + 1)
                            )
                        )
            else:  # no cell data
                num_nodes_per_cell = node_idcs.shape[1]

                form = '%d ' + '%d' % meshio_to_gmsh_type[cell_type] \
                    + ' 0 ' \
                    + ' ' + ' '.join(num_nodes_per_cell * ['%d']) \
                    + '\n'
                for k, c in enumerate(node_idcs):
                    fh.write(
                        form % (
                            (consecutive_index + k + 1,) +
                            tuple(c + 1)
                            )
                        )
            consecutive_index += data.shape[0]
        fh.write('$EndElements')

    return



# -*- coding: utf-8 -*-
#
'''
I/O for the OFF surface format, cf.
<https://en.wikipedia.org/wiki/OFF_(file_format)>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
from itertools import islice
import numpy


def read(filename):
    with open(filename) as f:
        points, cells = read_buffer(f)

    return points, cells, {}, {}, {}


def read_buffer(f):
    # assert that the first line reads `OFF`
    line = next(islice(f, 1))
    assert line.strip() == 'OFF'

    # fast forward to the next significant line
    while True:
        line = next(islice(f, 1))
        stripped = line.strip()
        if len(stripped) > 0 and stripped[0] != '#':
            break

    # This next line contains:
    # <number of vertices> <number of faces> <number of edges>
    # 2775 5558 0
    num_verts, num_faces, num_edges = stripped.split(' ')
    num_verts = int(num_verts)
    num_faces = int(num_faces)
    num_edges = int(num_edges)

    verts = numpy.empty((num_verts, 3), dtype=float)

    # read vertices
    k = 0
    while True:
        if k >= num_verts:
            break

        try:
            line = next(islice(f, 1))
        except StopIteration:
            break
        stripped = line.strip()
        # skip comments and empty lines
        if len(stripped) == 0 or stripped[0] == '#':
            continue

        x, y, z = stripped.split()
        verts[k] = [float(x), float(y), float(z)]
        k += 1

    # read cells
    triangles = []
    k = 0
    while True:
        if k >= num_faces:
            break

        try:
            line = next(islice(f, 1))
        except StopIteration:
            break

        stripped = line.strip()

        # skip comments and empty lines
        if len(stripped) == 0 or stripped[0] == '#':
            continue

        data = stripped.split()
        num_points = int(data[0])
        assert num_points == len(data) - 1

        if not num_points == 3:
            raise RuntimeError('Can only handle triangular faces')

        data = [int(data[1]), int(data[2]), int(data[3])]
        triangles.append(data)

    cells = {}
    if len(triangles):
        cells['triangle'] = numpy.array(triangles)

    return verts, cells


def write(
        filename,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None
        ):

    for key in cells:
        if key not in ['triangle']:
            raise RuntimeError('Can only deal with triangular faces')

    tri = cells['triangle']

    with open(filename, 'w') as fh:
        fh.write('OFF\n')
        fh.write('# Created by meshio\n\n')

        # counts
        fh.write('%d %d %d\n\n' % (len(points), len(tri), 0))

        # vertices
        numpy.savetxt(fh, points, '%e %e %e')

        # triangles
        data_with_label = numpy.c_[
                tri.shape[1] * numpy.ones(tri.shape[0]),
                tri
                ]
        numpy.savetxt(fh, data_with_label, '%d  %d %d %d')

    return



# -*- coding: utf-8 -*-
#
from . import dolfin_io
from . import h5m_io
from . import medit_io
from . import msh_io
from . import off_io
from . import permas_io
from . import vtk_io

from .__about__ import (
    __version__,
    __author__,
    __author_email__,
    __website__
    )

import pipdated
if pipdated.needs_checking(__name__):
    msg = pipdated.check(__name__, __version__)
    if msg:
        print(msg)


input_filetypes = [
        'exodus',
        'gmsh',
        'dolfin-xml',
        'medit',
        'permas',
        'moab',
        'off',
        'vtk',
        'vtu',
        'xdmf',
        ]

output_filetypes = [
        'exodus',
        'gmsh',
        'dolfin-xml',
        'medit',
        'permas',
        'moab',
        'off',
        'vtk-ascii',
        'vtk-binary',
        'vtu',
        'xdmf',
        ]

_extension_to_filetype = {
    '.e': 'exodus',
    '.ex2': 'exodus',
    '.exo': 'exodus',
    '.mesh': 'medit',
    '.msh': 'gmsh',
    '.xml': 'dolfin-xml',
    '.post': 'permas',
    '.post.gz': 'permas',
    '.dato': 'permas',
    '.dato.gz': 'permas',
    '.h5m': 'moab',
    '.off': 'off',
    '.vtu': 'vtu',
    '.vtk': 'vtk',
    '.xdmf': 'xdmf',
    '.xmf': 'xdmf',
    }


def read(filename, file_format=None):
    '''Reads an unstructured mesh with added data.

    :param filenames: The files to read from.
    :type filenames: str
    :returns mesh{2,3}d: The mesh data.
    :returns point_data: Point data read from file.
    :type point_data: dict
    :returns field_data: Field data read from file.
    :type field_data: dict
    '''
    import os

    # http://stackoverflow.com/questions/4843173/how-to-check-if-a-type-of-variable-is-string-in-python
    assert isinstance(filename, str)

    if not file_format:
        # deduct file format from extension
        _, extension = os.path.splitext(filename)
        file_format = _extension_to_filetype[extension]

    if file_format == 'gmsh':
        return msh_io.read(filename)
    elif file_format == 'medit':
        return medit_io.read(filename)
    elif file_format == 'dolfin-xml':
        return dolfin_io.read(filename)
    elif file_format == 'permas':
        return permas_io.read(filename)
    elif file_format == 'moab':
        return h5m_io.read(filename)
    elif file_format == 'off':
        return off_io.read(filename)
    elif file_format == 'vtu':
        return vtk_io.read('vtu', filename)
    elif file_format == 'vtk':
        return vtk_io.read('vtk', filename)
    elif file_format == 'xdmf':
        return vtk_io.read('xdmf', filename)
    elif file_format == 'exodus':
        return vtk_io.read('exodus', filename)
    else:
        raise RuntimeError(
            'Unknown file format \'%s\' of file \'%s\'.' %
            (file_format, filename)
            )


def write(filename,
          points,
          cells,
          file_format=None,
          point_data=None,
          cell_data=None,
          field_data=None
          ):
    '''Writes mesh together with data to a file.

    :params filename: File to write to.
    :type filename: str

    :params point_data: Named additional point data to write to the file.
    :type point_data: dict
    '''
    import os

    if not file_format:
        # deduct file format from extension
        _, extension = os.path.splitext(filename)
        file_format = _extension_to_filetype[extension]

    # check cells for sanity
    if 'vertex' in cells:
        assert cells['vertex'].shape[1] == 1
    if 'line' in cells:
        assert cells['line'].shape[1] == 2
    if 'triangle' in cells:
        assert cells['triangle'].shape[1] == 3
    if 'quad' in cells:
        assert cells['quad'].shape[1] == 4
    if 'tetra' in cells:
        assert cells['tetra'].shape[1] == 4
    if 'hexahedron' in cells:
        assert cells['hexahedron'].shape[1] == 8
    if 'wedge' in cells:
        assert cells['wedge'].shape[1] == 6
    if 'pyramid' in cells:
        assert cells['pyramid'].shape[1] == 5

    if file_format == 'moab':
        h5m_io.write(
            filename, points, cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data
            )
    elif file_format == 'gmsh':
        msh_io.write(
            filename, points, cells,
            point_data=point_data,
            cell_data=cell_data
            )
    elif file_format == 'medit':
        medit_io.write(filename, points, cells)
    elif file_format == 'dolfin-xml':
        dolfin_io.write(filename, points, cells)
    elif file_format == 'off':
        off_io.write(filename, points, cells)
    elif file_format == 'permas':
        permas_io.write(filename, points, cells)
    elif file_format == 'vtu':  # vtk xml format
        vtk_io.write(
            'vtu', filename, points, cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data
            )
    elif file_format in ['vtk', 'vtk-ascii']:
        vtk_io.write(
            'vtk-ascii', filename,
            points, cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data
            )
    elif file_format == 'vtk-binary':
        vtk_io.write(
            'vtk-binary', filename,
            points, cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data
            )
    elif file_format == 'xdmf':  # XDMF
        vtk_io.write(
            'xdmf', filename, points, cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data
            )
    elif file_format == 'exodus':  # exodus ii format
        vtk_io.write(
            'exodus', filename, points, cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data
            )
    else:
        raise RuntimeError(
            'unknown file format \'%s\' of \'%s\'.' % (file_format, filename)
            )
    return



# -*- coding: utf-8 -*-
#
import meshio

import numpy
import os
import pytest

tri_mesh = {
        'points': numpy.array([
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0]
            ]),
        'cells': {
            'triangle': numpy.array([
                [0, 1, 2],
                [0, 2, 3]
                ])
            },
        }

quad_mesh = {
        'points': numpy.array([
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [2.0, 0.0, 0.0],
            [2.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0]
            ]),
        'cells': {
            'quad': numpy.array([
                [0, 1, 4, 5],
                [1, 2, 3, 4]
                ])
            },
        }

tri_quad_mesh = {
        'points': numpy.array([
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [2.0, 0.0, 0.0],
            [2.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0]
            ]),
        'cells':  {
            'triangle': numpy.array([
                [0, 1, 4],
                [0, 4, 5]
                ]),
            'quad': numpy.array([
                [1, 2, 3, 4]
                ])
            }
        }

tet_mesh = {
        'points': numpy.array([
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.5, 0.5, 0.5],
            ]),
        'cells': {
            'tetra': numpy.array([
                [0, 1, 2, 4],
                [0, 2, 3, 4]
                ])
            },
        }


def _clone(mesh):
    mesh2 = {
        'points': numpy.copy(mesh['points'])
        }
    mesh2['cells'] = {}
    for key, data in mesh['cells'].iteritems():
        mesh2['cells'][key] = numpy.copy(data)
    return mesh2


def _add_point_data(mesh, dim):
    numpy.random.seed(0)
    mesh2 = _clone(mesh)

    # Don't use default float64 here because of a bug in VTK that prevents
    # actual double data from being written out, cf.
    # <http://www.vtk.org/Bug/view.php?id=15889>.
    if dim == 1:
        data = numpy.random.rand(len(mesh['points'])).astype('float32')
    else:
        data = numpy.random.rand(len(mesh['points']), dim).astype('float32')

    mesh2['point_data'] = {'a': data}
    return mesh2


def _add_cell_data(mesh, dim):
    mesh2 = _clone(mesh)
    numpy.random.seed(0)
    cell_data = {}
    for cell_type in mesh['cells']:
        num_cells = len(mesh['cells'][cell_type])
        if dim == 1:
            cell_data[cell_type] = {
                'b': numpy.random.rand(num_cells)
                }
        else:
            cell_data[cell_type] = {
                'b': numpy.random.rand(num_cells, dim)
                }

    mesh2['cell_data'] = cell_data
    return mesh2


@pytest.mark.parametrize('extension, meshes', [
    ('dato', [tri_mesh, quad_mesh, tri_quad_mesh, tet_mesh]),
    # ('e', [
    #     tri_mesh,
    #     _add_point_data(tri_mesh, 2),
    #     _add_point_data(tri_mesh, 3)
    #     ]),
    ('h5m', [tri_mesh, tet_mesh]),
    ('msh', [tri_mesh, quad_mesh, tri_quad_mesh, tet_mesh]),
    ('mesh', [tri_mesh, quad_mesh, tri_quad_mesh, tet_mesh]),
    ('off', [tri_mesh]),
    # These tests needed to be disabled since travis-ci only offers trusty with
    # the buggy VTK 6.0.
    # TODO enable once we can use a more recent version of VTK
    # ('vtk', [
    #      tri_mesh,
    #      quad_mesh,
    #      tri_quad_mesh,
    #      tet_mesh,
    #      _add_point_data(tri_mesh, 1),
    #      _add_point_data(tri_mesh, 2),
    #      _add_point_data(tri_mesh, 3),
    #      _add_cell_data(tri_mesh, 1),
    #      _add_cell_data(tri_mesh, 2),
    #      _add_cell_data(tri_mesh, 3),
    #      ]),
    # ('vtu', [
    #      tri_mesh,
    #      quad_mesh,
    #      tri_quad_mesh,
    #      tet_mesh,
    #      _add_point_data(tri_mesh, 1),
    #      _add_point_data(tri_mesh, 2),
    #      _add_point_data(tri_mesh, 3),
    #      _add_cell_data(tri_mesh, 1),
    #      _add_cell_data(tri_mesh, 2),
    #      _add_cell_data(tri_mesh, 3),
    #      ]),
    #
    # 2016-04-27: Temporarily disabled due to vtkXdmfWriter not being available
    #             through VTK
    # ('xdmf', [
    #     tri_mesh,
    #     quad_mesh,
    #     tet_mesh
    #     # The two following tests pass, but errors are emitted on the
    #     # console.
    #     # _add_point_data(tri_mesh, 1),
    #     # _add_cell_data(tri_mesh, 1)
    #     ]),
    ('xml', [tri_mesh, tet_mesh]),
    ])
def test_io(extension, meshes):
    filename = 'test.' + extension
    for mesh in meshes:
        _write_read(filename, mesh)


def _write_read(filename, mesh):
    '''Write and read a file, and make sure the data is the same as before.
    '''
    try:
        input_point_data = mesh['point_data']
    except KeyError:
        input_point_data = {}

    try:
        input_cell_data = mesh['cell_data']
    except KeyError:
        input_cell_data = {}

    meshio.write(
        filename,
        mesh['points'], mesh['cells'],
        point_data=input_point_data,
        cell_data=input_cell_data
        )
    points, cells, point_data, cell_data, _ = meshio.read(filename)

    # Numpy's array_equal is too strict here, cf.
    # <https://mail.scipy.org/pipermail/numpy-discussion/2015-December/074410.html>.
    # Use allclose.

    # We cannot compare the exact rows here since the order of the points might
    # have changes. Just compare the sums
    assert numpy.allclose(mesh['points'], points)

    for cell_type, data in mesh['cells'].items():
        assert numpy.allclose(data, cells[cell_type])
    for key, data in input_point_data.items():
        assert numpy.allclose(data, point_data[key])
    for cell_type, cell_type_data in input_cell_data.items():
        for key, data in cell_type_data.iteritems():
            assert numpy.allclose(data, cell_data[cell_type][key])

    os.remove(filename)

    return



# -*- coding: utf-8 -*-
#
'''
I/O for DOLFIN's XML format, cf.
<https://people.sc.fsu.edu/~jburkardt/data/dolfin_xml/dolfin_xml.html>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
import numpy
import warnings


def read(filename):
    from lxml import etree as ET

    tree = ET.parse(filename)
    root = tree.getroot()
    mesh = root.getchildren()[0]
    assert mesh.tag == 'mesh'

    dolfin_to_meshio_type = {
        'triangle': ('triangle', 3),
        'tetrahedron': ('tetra', 4),
        }

    cell_type, npc = dolfin_to_meshio_type[mesh.attrib['celltype']]

    is_2d = mesh.attrib['dim'] == '2'
    if not is_2d:
        assert mesh.attrib['dim'] == '3'

    points = None
    cells = {
        cell_type: None
        }

    for child in mesh.getchildren():
        if child.tag == 'vertices':
            num_verts = int(child.attrib['size'])
            points = numpy.empty((num_verts, 3))
            for vert in child.getchildren():
                assert vert.tag == 'vertex'
                idx = int(vert.attrib['index'])
                points[idx, 0] = vert.attrib['x']
                points[idx, 1] = vert.attrib['y']
                if is_2d:
                    points[idx, 2] = 0.0
                else:
                    points[idx, 2] = vert.attrib['z']

        elif child.tag == 'cells':
            num_cells = int(child.attrib['size'])
            cells[cell_type] = numpy.empty((num_cells, npc), dtype=int)
            for cell in child.getchildren():
                assert(dolfin_to_meshio_type[cell.tag][0] == cell_type)
                idx = int(cell.attrib['index'])
                for k in range(npc):
                    cells[cell_type][idx, k] = cell.attrib['v%s' % k]

        else:
            raise RuntimeError('Unknown entry \'%s\'.' % child.tag)

    point_data = {}
    cell_data = {}
    field_data = {}
    return points, cells, point_data, cell_data, field_data


def write(
        filename,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None
        ):
    from lxml import etree as ET

    if point_data is None:
        point_data = {}
    if cell_data is None:
        cell_data = {}
    if field_data is None:
        field_data = {}

    dolfin = ET.Element(
        'dolfin',
        nsmap={'dolfin': 'http://fenicsproject.org/'}
        )

    meshio_to_dolfin_type = {
            'triangle': 'triangle',
            'tetra': 'tetrahedron',
            }

    if 'tetra' in cells:
        stripped_cells = {'tetra': cells['tetra']}
        cell_type = 'tetra'
    elif 'triangle' in cells:
        stripped_cells = {'triangle': cells['triangle']}
        cell_type = 'triangle'
    else:
        raise RuntimeError(
          'Dolfin XML can only deal with triangle or tetra. '
          'The input data contains only ' + ', '.join(cells.keys()) + '.'
          )

    if len(cells) > 1:
        discarded_cells = cells.keys()
        discarded_cells.remove(cell_type)
        warnings.warn(
          'DOLFIN XML can only handle one cell type at a time. '
          'Using ' + cell_type +
          ', discarding ' + ', '.join(discarded_cells) +
          '.'
          )

    if all(points[:, 2] == 0):
        dim = '2'
    else:
        dim = '3'

    mesh = ET.SubElement(
            dolfin,
            'mesh',
            celltype=meshio_to_dolfin_type[cell_type],
            dim=dim
            )
    vertices = ET.SubElement(mesh, 'vertices', size=str(len(points)))
    for k, point in enumerate(points):
        ET.SubElement(
            vertices,
            'vertex',
            index=str(k),
            x=str(point[0]), y=str(point[1]), z=str(point[2])
            )

    num_cells = 0
    for cls in stripped_cells.values():
        num_cells += len(cls)

    xcells = ET.SubElement(mesh, 'cells', size=str(num_cells))
    idx = 0
    for cell_type, cls in stripped_cells.iteritems():
        for cell in cls:
            cell_entry = ET.SubElement(
                xcells,
                meshio_to_dolfin_type[cell_type],
                index=str(idx)
                )
            for k, c in enumerate(cell):
                cell_entry.attrib['v%d' % k] = str(c)
            idx += 1

    tree = ET.ElementTree(dolfin)
    tree.write(filename, pretty_print=True)
    return



# -*- coding: utf-8 -*-
#
'''
I/O for PERMAS dat format, cf.

.. moduleauthor:: Nils Wagner
'''

import gzip
import numpy
import re
from .__about__ import __version__, __website__, __author__


def read(filename):
    '''Reads a (compressed) PERMAS dato or post file.
    '''
    # The format is specified at
    # <http://www.intes.de>.
    if filename.endswith('post.gz'):
        f = gzip.open(filename, 'r')
    elif filename.endswith('dato.gz'):
        f = gzip.open(filename, 'r')
    elif filename.endswith('dato'):
        f = open(filename, 'r')
    elif filename.endswith('post'):
        f = open(filename, 'r')
    else:
        print('Unsupported file format')

    cells = {}
    meshio_to_permas_type = {
        'line': (2, 'PLOTL2'),
        'triangle': (3, 'TRIA3'),
        'quad': (4, 'QUAD4'),
        'tetra': (4, 'TET4'),
        'hexahedron': (8, 'HEXE8'),
        'wedge': (6, 'PENTA6'),
        'pyramid': (5, 'PYRA5')
        }

    while True:
        line = f.readline()
        if not line or re.search('\$END STRUCTURE', line):
            break
        for meshio_type, permas_ele in meshio_to_permas_type.iteritems():
            num_nodes = permas_ele[0]
            permas_type = permas_ele[1]

            if re.search('\$ELEMENT TYPE = %s' % permas_type, line):
                while True:
                    line = f.readline()
                    if not line or line.startswith('!'):
                        break
                    data = numpy.array(line.split(), dtype=int)
                    if meshio_type in cells:
                        cells[meshio_type].append(data[-num_nodes:])
                    else:
                        cells[meshio_type] = [data[-num_nodes:]]

        if re.search('\$COOR', line):
            points = []
            while True:
                line = f.readline()
                if not line or line.startswith('!'):
                    break
                for r in numpy.array(line.split(), dtype=float)[1:]:
                    points.append(r)
    points = numpy.reshape(points, newshape=(len(points)/3, 3))
    for key in cells:
        # Subtract one to account for the fact that python indices
        # are 0-based.
        cells[key] = numpy.array(cells[key], dtype=int) - 1

    return points, cells, {}, {}, {}


def write(
        filename,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None
        ):
    '''Writes PERMAS dat files, cf.
    http://www.intes.de # PERMAS-ASCII-file-format
    '''
    if point_data is None:
        point_data = {}
    if cell_data is None:
        cell_data = {}
    if field_data is None:
        field_data = {}

    with open(filename, 'w') as fh:
        fh.write('!\n')
        fh.write('! Author of meshio: %s\n' % __author__)
        fh.write('! File written by meshio version %s\n' % __version__)
        fh.write('! Further information available at %s\n' % __website__)
        fh.write('!\n')
        fh.write('$ENTER COMPONENT NAME = DFLT_COMP DOFTYPE = DISP MATH\n')
        fh.write('! \n')
        fh.write('    $SITUATION NAME = REAL_MODES\n')
        fh.write('        DFLT_COMP SYSTEM = NSV CONSTRAINTS = SPCVAR_1 ! LOADING = LOADVAR_1\n')
        fh.write('    $END SITUATION\n')
        fh.write('! \n')
        fh.write('    $STRUCTURE\n')
        fh.write('! \n')

        # Write nodes
        fh.write('        $COOR NSET = ALL_NODES\n')
        for k, x in enumerate(points):
            fh.write(
                '        %8d %+.15f %+.15f %+.15f\n' %
                (k+1, x[0], x[1], x[2])
                )

        meshio_to_permas_type = {
            'line': (2, 'PLOTL2'),
            'triangle': (3, 'TRIA3'),
            'quad': (4, 'QUAD4'),
            'tetra': (4, 'TET4'),
            'hexahedron': (8, 'HEXE8'),
            'wedge': (6, 'PENTA6'),
            'pyramid': (5, 'PYRA5')
            }

        #
        # Avoid non-unique element numbers in case of multiple element types by
        # num_ele !!!
        #
        num_ele = 0

        for meshio_type, cell in cells.iteritems():
            numcells, num_local_nodes = cell.shape
            permas_type = meshio_to_permas_type[meshio_type]
            fh.write('!\n')
            fh.write(
                '        $ELEMENT TYPE = %s ESET = %s\n' %
                (permas_type[1], permas_type[1])
                )
            for k, c in enumerate(cell):
                form = '        %8d ' + \
                    ' '.join(num_local_nodes * ['%8d']) + \
                    '\n'
                fh.write(form % ((k+num_ele+1,) + tuple(c + 1)))
            num_ele += numcells

        fh.write('!\n')
        fh.write('    $END STRUCTURE\n')
        fh.write('!\n')
        elem_3D = ['HEXE8', 'TET4', 'PENTA6', 'PYRA5']
        elem_2D = ['TRIA3', 'QUAD4']
        fh.write('    $SYSTEM NAME = NSV\n')
        fh.write('!\n')
        fh.write('        $ELPROP\n')
        for meshio_type, cell in cells.iteritems():
            permas_type = meshio_to_permas_type[meshio_type]
            if permas_type[1] in elem_3D:
                fh.write(
                    '            %s MATERIAL = DUMMY_MATERIAL\n' %
                    permas_type[1]
                    )
            elif permas_type[1] in elem_2D:
                fh.write('            %s GEODAT = GD_%s MATERIAL = DUMMY_MATERIAL\n' %(permas_type[1],permas_type[1]))
            else:
                pass
        fh.write('!\n')
        fh.write('        $GEODAT SHELL  CONT = THICK  NODES = ALL\n')
        for meshio_type, cell in cells.iteritems():
            permas_type = meshio_to_permas_type[meshio_type]
            if permas_type[1] in elem_2D:
                fh.write('            GD_%s 1.0\n' % permas_type[1])
        fh.write('!\n')
        fh.write('    $END SYSTEM\n')
        fh.write('!\n')
        fh.write('    $CONSTRAINTS NAME = SPCVAR_1\n')
        fh.write('    $END CONSTRAINTS\n')
        fh.write('!\n')
        fh.write('    $LOADING NAME = LOADVAR_1\n')
        fh.write('    $END LOADING\n')
        fh.write('!\n')
        fh.write('$EXIT COMPONENT\n')
        fh.write('!\n')
        fh.write('$ENTER MATERIAL\n')
        fh.write('!\n')
        fh.write('    $MATERIAL NAME = DUMMY_MATERIAL TYPE = ISO\n')
        fh.write('!\n')
        fh.write('        $ELASTIC  GENERAL  INPUT = DATA\n')
        fh.write('            0.0 0.0\n')
        fh.write('!\n')
        fh.write('        $DENSITY  GENERAL  INPUT = DATA\n')
        fh.write('            0.0\n')
        fh.write('!\n')
        fh.write('        $THERMEXP  GENERAL  INPUT = DATA\n')
        fh.write('            0.0\n')
        fh.write('!\n')
        fh.write('    $END MATERIAL\n')
        fh.write('!\n')
        fh.write('$EXIT MATERIAL\n')
        fh.write('!\n')
        fh.write('$FIN\n')
    return



# -*- coding: utf-8 -*-
#
import os
from distutils.core import setup
import codecs

# https://packaging.python.org/single_source_version/
base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, 'meshio', '__about__.py')) as f:
    exec(f.read(), about)


def read(fname):
    try:
        content = codecs.open(
            os.path.join(os.path.dirname(__file__), fname),
            encoding='utf-8'
            ).read()
    except Exception:
        content = ''
    return content


setup(
    name='meshio',
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    packages=['meshio'],
    description='I/O for various mesh formats',
    long_description=read('README.rst'),
    url='https://github.com/nschloe/meshio',
    download_url='https://pypi.python.org/pypi/meshio',
    license='License :: OSI Approved :: MIT License',
    platforms='any',
    install_requires=[
        'h5py',
        'lxml',
        'numpy',
        'pipdated',
        ],
    scripts=[
      'tools/meshio-convert',
      ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Topic :: Scientific/Engineering'
        ]
    )
